use eyre::Result;
use once_cell::sync::Lazy;
use reqwest::{Response, StatusCode};
use rsa::RsaPrivateKey;
use serde::{de::DeserializeOwned, Serialize};

use crate::{
    signature::{build_signature, format_date},
    FoxError,
};

static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

static CLIENT: Lazy<reqwest::Client> = Lazy::new(|| {
    reqwest::Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()
        .unwrap()
});

pub fn is_valid_domain(domain: &str) -> bool {
    addr::parse_domain_name(domain).is_ok()
}

pub async fn get<R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: String,
    host: String,
    path: String,
    user_id: Option<String>,
) -> Result<R> {
    let (signature, date) = build_signature(
        private_key,
        host.clone(),
        path.clone(),
        None,
        user_id.clone(),
    );

    let mut req = CLIENT
        .get(format!("https://{host}{path}"))
        .header("Date", format_date(date))
        .header("X-Foxchat-Signature", signature)
        .header("X-Foxchat-Server", self_domain);

    req = if let Some(id) = user_id {
        req.header("X-Foxchat-User", id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await
}

pub async fn post<T: Serialize, R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: String,
    host: String,
    path: String,
    user_id: Option<String>,
    body: &T,
) -> Result<R> {
    let body = serde_json::to_string(body)?;

    let (signature, date) = build_signature(
        private_key,
        host.clone(),
        path.clone(),
        Some(body.len()),
        user_id.clone(),
    );

    let mut req = CLIENT
        .post(format!("https://{host}{path}"))
        .header("Date", format_date(date))
        .header("X-Foxchat-Signature", signature)
        .header("X-Foxchat-Server", self_domain)
        .header("Content-Type", "application/json; charset=utf-8")
        .header("Content-Length", body.len().to_string())
        .body(body);

    req = if let Some(id) = user_id {
        req.header("X-Foxchat-User", id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await
}

async fn handle_response<R: DeserializeOwned>(resp: Response) -> Result<R> {
    if resp.status() != StatusCode::OK {
        return Err(FoxError::ResponseNotOk.into());
    }

    let parsed = resp.json::<R>().await?;
    Ok(parsed)
}
